import functools
import ipaddress
import re

from django.conf import settings
from django.db import models
from django.http import HttpRequest


class invertedset(frozenset):
    def __contains__(self, item):
        return not super().__contains__(item)


class RULE_ACTION(models.TextChoices):
    allow = ("a", "allow")
    deny = ("b", "deny")
    disabled = ("c", "disabled")
    only_ratelimit = ("d", "only apply ratelimit")


def get_default_interval():
    return 24 * 60 * 60


@functools.lru_cache(maxsize=512)
def parse_ipaddress(inp):
    ip = ipaddress.ip_address(inp)
    mapped = getattr(ip, "ipv4_mapped", None)
    if mapped:
        ip = mapped
    return ip


@functools.lru_cache(maxsize=512)
def parse_ipnetwork(inp):
    return ipaddress.ip_network(inp, strict=False)


@functools.lru_cache(maxsize=1)
def get_TRUSTED_PROXY() -> frozenset:
    setting = getattr(settings, "IPRESTRICT_TRUSTED_PROXIES", None)
    if not setting:
        setting = getattr(settings, "RATELIMIT_TRUSTED_PROXIES", ["unix"])
    if setting == "all":
        return invertedset()
    else:
        return frozenset(setting)


def get_FALLBACK() -> str:
    setting = getattr(settings, "IPRESTRICT_TESTCLIENT_FALLBACK", None)
    if not setting:
        setting = getattr(settings, "RATELIMIT_TESTCLIENT_FALLBACK", "::1")
    return setting


_forwarded_regex = re.compile(r'for="?([^";, ]+)', re.IGNORECASE)
_http_x_forwarded_regex = re.compile(r'[ "]*([^";, ]+)')
_ip6_port_cleanup_regex = re.compile(r"(?<=\]):[0-9]+$")
_ip4_port_cleanup_regex = re.compile(r":[0-9]+$")


def get_ip(request: HttpRequest):
    client_ip = request.META.get("REMOTE_ADDR", "") or "unix"
    if client_ip in get_TRUSTED_PROXY():
        try:
            ip_matches = _forwarded_regex.search(request.META["HTTP_FORWARDED"])
            client_ip = ip_matches[1]
        except KeyError:
            try:
                ip_matches = _http_x_forwarded_regex.search(
                    request.META["HTTP_X_FORWARDED_FOR"]
                )
                client_ip = ip_matches[1]
            except KeyError:
                pass
    if client_ip == "testclient":  # starlite test client
        client_ip = get_FALLBACK()
    if client_ip in {"unix", "invalid"}:
        raise ValueError("Could not determinate ip address")
    if "." in client_ip and client_ip.count(":") <= 1:
        client_ip = _ip4_port_cleanup_regex.sub("", client_ip)
    else:
        client_ip = _ip6_port_cleanup_regex.sub("", client_ip).strip("[]")

    return client_ip


def get_default_action(action=None):
    if action:
        return action
    action = RULE_ACTION[getattr(settings, "IPRESTRICT_DEFAULT_ACTION", "allow")]
    assert action != RULE_ACTION.disabled, "disabled is not a valid default action"
    return action.value


class LockoutException(Exception):
    pass



# Generated by Django 5.0 on 2023-12-31 11:29

import django.db.models.deletion
from django.db import migrations, models

import django_fast_iprestrict.validators


class Migration(migrations.Migration):
    dependencies = [
        ("django_fast_iprestrict", "0004_cleanup_rule"),
    ]

    operations = [
        migrations.AlterField(
            model_name="rule",
            name="action",
            field=models.CharField(
                choices=[
                    ("a", "allow"),
                    ("b", "deny"),
                    ("c", "disabled"),
                    ("d", "only apply ratelimit"),
                ],
                default="c",
                max_length=1,
            ),
        ),
        migrations.AlterField(
            model_name="rulesource",
            name="generator_fn",
            field=models.CharField(
                max_length=200,
                validators=[django_fast_iprestrict.validators.validate_generator_fn],
            ),
        ),
        migrations.CreateModel(
            name="RuleRatelimit",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "key",
                    models.CharField(
                        max_length=200,
                        validators=[
                            django_fast_iprestrict.validators.validate_ratelimit_key
                        ],
                    ),
                ),
                ("group", models.CharField(max_length=50)),
                (
                    "decorate_name",
                    models.CharField(blank=True, default="ratelimit", max_length=50),
                ),
                (
                    "rate",
                    models.CharField(
                        max_length=10,
                        validators=[django_fast_iprestrict.validators.validate_rate],
                    ),
                ),
                ("block", models.BooleanField(blank=True, default=False)),
                ("wait", models.BooleanField(blank=True, default=False)),
                ("is_active", models.BooleanField(blank=True, default=True)),
                (
                    "rule",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="ratelimits",
                        to="django_fast_iprestrict.rule",
                    ),
                ),
            ],
        ),
    ]



from posixpath import dirname

from django.contrib import admin
from django.core.exceptions import PermissionDenied
from django.http import (
    HttpRequest,
    HttpResponse,
    HttpResponseBadRequest,
    HttpResponseRedirect,
)
from django.template.response import TemplateResponse
from django.urls import path
from django.utils.html import format_html

from .models import Rule, RuleNetwork, RulePath, RuleRatelimit, RuleSource
from .utils import RULE_ACTION, LockoutException, get_ip, parse_ipaddress

try:
    import django_fast_ratelimit as ratelimit
except ImportError:
    ratelimit = None


# Register your models here.


class RuleSubMixin:
    def get_urls(self):
        return [
            path("<str:object_id>/change/", self.redirect_change),
            *super().get_urls(),
        ]

    def redirect_change(self, request, object_id):
        rulep = RulePath.objects.get(id=object_id)
        return HttpResponseRedirect(f"../../../rule/{rulep.rule_id}/change/")

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False


class ExtraOnlyOnInitialMixin:
    extra = 1

    def get_extra(self, request, obj=None, **kwargs):
        return self.extra if not obj else 0


class RulePathInlineAdmin(ExtraOnlyOnInitialMixin, admin.StackedInline):
    model = RulePath


class RuleRatelimitInlineAdmin(ExtraOnlyOnInitialMixin, admin.StackedInline):
    model = RuleRatelimit


class RuleNetworkInlineAdmin(ExtraOnlyOnInitialMixin, admin.TabularInline):
    model = RuleNetwork


class RuleSourceInlineAdmin(ExtraOnlyOnInitialMixin, admin.TabularInline):
    model = RuleSource


# reuse the existing form to prevent form nesting
_position_template = """
<div style="display:flex; gap: 16px" >
    <button
        style="border-radius: 30%;"
        name="rule_move_direction"
        title="start"
        value="start"
        formaction="{object_id}/rule_move/"
    >
        &#x21C8;
    </button>
    <button
        style="border-radius: 30%;"
        name="rule_move_direction"
        title="up" value="up"
        formaction="{object_id}/rule_move/"
    >
        &#x2191;
    </button>
    <button
        style="border-radius: 30%;"
        name="rule_move_direction"
        title="down" value="down"
        formaction="{object_id}/rule_move/"
    >
        &#x2193;
    </button>
    <button
        style="border-radius: 30%;"
        name="rule_move_direction"
        title="end"
        value="end"
        formaction="{object_id}/rule_move/"
    >
        &#x21CA;
    </button>
</div>
"""


@admin.register(Rule)
class RuleAdmin(admin.ModelAdmin):
    list_display = ("position_short", "position_buttons", "action", "name")
    list_display_links = ("position_short",)
    list_editable = ("name", "action")
    ordering = ("position",)
    fields = ["name", "action"]
    inlines = [
        RuleNetworkInlineAdmin,
        RuleSourceInlineAdmin,
        RulePathInlineAdmin,
        RuleRatelimitInlineAdmin,
    ]

    @admin.display(description="")
    def position_short(self, obj):
        return obj.position

    @admin.display(ordering="position")
    def position_buttons(self, obj):
        return format_html(
            _position_template,
            object_id=obj.id,
        )

    def get_urls(self):
        return [
            path("test_rules/", self.test_rules),
            path(
                "simulate_rules<path:test_path>",
                self.simulate_rules,
            ),
            path("<str:object_id>/rule_move/", self.rule_move),
            *super().get_urls(),
        ]

    def rule_move(self, request: HttpRequest, object_id):
        direction = request.POST["rule_move_direction"]
        parent_path = dirname(request.path.rstrip("/"))
        try:
            if direction == "up":
                self.model.objects.position_up(
                    object_id, ip=get_ip(request), path=parent_path
                )
            elif direction == "down":
                self.model.objects.position_down(
                    object_id, ip=get_ip(request), path=parent_path
                )
            elif direction == "start":
                self.model.objects.position_start(
                    object_id, ip=get_ip(request), path=parent_path
                )
            elif direction == "end":
                self.model.objects.position_end(
                    object_id, ip=get_ip(request), path=parent_path
                )
            else:
                return HttpResponseBadRequest()
        except LockoutException:
            context = {**self.admin_site.each_context(request), "link_back": "../../"}
            return TemplateResponse(
                request,
                "admin/django_fast_iprestrict/lockout_prevented.html",
                context=context,
                status=400,
            )
        return HttpResponseRedirect("../../")

    def simulate_rules(self, request, test_path):
        rule_id, action, is_catch_all, ratelimits = RulePath.objects.match_ip_and_path(
            get_ip(request), test_path
        )

        if ratelimit:
            for rdict in ratelimits:
                r = ratelimit.get_ratelimit(
                    request=request,
                    action=ratelimit.Action.PEEK,
                    group=rdict["group"],
                    key=rdict["key"],
                    rate=rdict["rate"],
                )
                r.decorate_object(
                    request, name=rdict["decorate_name"], block=rdict["block"]
                )
        if action == RULE_ACTION.deny:
            raise PermissionDenied()

        return HttpResponse(f"accessed: {test_path}")

    def test_rules(self, request):
        from django.contrib.messages import ERROR, INFO, SUCCESS, WARNING

        test_ip = request.POST.get("test_ip", None)
        if not test_ip:
            test_ip = get_ip(request)

        test_path = request.POST.get("test_path", None) or ""
        if test_path:
            rule_id = RulePath.objects.match_ip_and_path(test_ip, test_path)[0]
            self.message_user(
                request,
                f"Parameters: ip: {test_ip} (parsed: {parse_ipaddress(test_ip)}), path: {test_path}",
                level=INFO,
            )
        else:
            rule_id = Rule.objects.match_ip(test_ip)[0]
            self.message_user(
                request,
                f"Parameters: ip: {test_ip} (parsed: {parse_ipaddress(test_ip)})",
                level=INFO,
            )
        if rule_id:
            rule = Rule.objects.get(id=rule_id)
            self.message_user(
                request,
                f"Matched rule: {rule.name}, action: {RULE_ACTION(rule.action).label}",
                level=SUCCESS if rule.action == RULE_ACTION.allow.value else WARNING,
            )
        else:
            self.message_user(request, "No rule matched", level=ERROR)

        return HttpResponseRedirect("../")

    def changeform_view(self, request, object_id=None, form_url="", extra_context=None):
        try:
            return super().changeform_view(
                request,
                object_id=object_id,
                form_url=form_url,
                extra_context=extra_context,
            )
        except LockoutException:
            context = {**self.admin_site.each_context(request), "link_back": "./"}
            return TemplateResponse(
                request,
                "admin/django_fast_iprestrict/lockout_prevented.html",
                context=context,
                status=400,
            )

    def changelist_view(self, request, extra_context=None):
        try:
            return super().changelist_view(
                request,
                extra_context=extra_context,
            )
        except LockoutException:
            context = {**self.admin_site.each_context(request), "link_back": "./"}
            return TemplateResponse(
                request,
                "admin/django_fast_iprestrict/lockout_prevented.html",
                context=context,
                status=400,
            )

    def save_model(self, request, obj, form, change):
        obj._trigger_cleanup = False
        super().save_model(request, obj, form, change)

    def delete_model(self, request, obj):
        obj._trigger_cleanup = False
        super().delete_model(request, obj)
        Rule.objects.position_cleanup(ip=get_ip(request), path=request.path)

    def delete_queryset(self, request, queryset):
        super().delete_queryset(request, queryset)
        parent_path = dirname(request.path.rstrip("/"))

        Rule.objects.position_cleanup(ip=get_ip(request), path=parent_path)

    def save_related(self, request, form, formsets, change):
        super().save_related(request, form, formsets, change)
        Rule.objects.position_cleanup(ip=get_ip(request), path=request.path)


@admin.register(RulePath)
class RulePathAdmin(RuleSubMixin, admin.ModelAdmin):
    list_display = ("rule", "path", "is_regex", "is_active")
    ordering = ("rule", "id")


@admin.register(RuleNetwork)
class RuleNetworkAdmin(RuleSubMixin, admin.ModelAdmin):
    list_display = ("rule", "network", "is_active")
    ordering = ("rule", "id")


@admin.register(RuleSource)
class RuleSourceAdmin(RuleSubMixin, admin.ModelAdmin):
    list_display = ("rule", "generator_fn", "is_active")
    ordering = ("rule", "id")



__all__ = ["apply_iprestrict"]

from functools import partial, singledispatch

from django.http import HttpRequest

from .utils import RULE_ACTION, get_default_action, get_ip

try:
    import django_fast_ratelimit as ratelimit
except ImportError:
    ratelimit = None


@singledispatch
def apply_iprestrict(
    request: HttpRequest, group, ignore_pathes=False, require_rule=False
):
    from .models import Rule, RulePath

    rule = Rule.objects.filter(name=group).first()
    if not rule:
        if require_rule:
            return 1
        return int(get_default_action() == RULE_ACTION.deny)
    with_path = False
    ip = get_ip(request)
    if not ignore_pathes:
        with_path = rule.id in RulePath.objects.path_matchers()
    if with_path:
        action, _, ratelimits = RulePath.objects.match_ip_and_path(
            ip=ip, path=request.path
        )[1:]
    else:
        action, _, ratelimits = rule.match_ip(ip=ip)[1:]
    if ratelimit:
        for rdict in ratelimits:
            r = ratelimit.get_ratelimit(
                request=request,
                action=ratelimit.Action.INCREASE,
                group=rdict["group"],
                key=rdict["key"],
                rate=rdict["rate"],
            )
            r.decorate_object(
                request, name=rdict["decorate_name"], block=rdict["block"]
            )

    if action == RULE_ACTION.deny.value:
        return 1

    return 0


@apply_iprestrict.register
def _(arg: str = ""):
    args = arg.split(",")
    return partial(
        apply_iprestrict,
        ignore_pathes="ignore_pathes" in args,
        require_rule="require_rule" in args,
    )



from django.core.management.base import BaseCommand

from ...misc import RULE_ACTION
from ...models import Rule


class Command(BaseCommand):
    help = "Emergency disable all ratelimits and clear caches for reallowing login"

    def handle(self, **kwargs):
        Rule.objects.filter(action=RULE_ACTION.deny.value).update(
            action=RULE_ACTION.disabled.value
        )
        # clearing local_caches doesn't help, different proccess



# Generated by Django 5.0 on 2023-12-28 03:24

import django.db.models.deletion
from django.db import migrations, models

import django_fast_iprestrict.models
import django_fast_iprestrict.validators


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Rule",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "position",
                    models.PositiveIntegerField(
                        blank=True,
                        default=django_fast_iprestrict.models.get_default_position,
                    ),
                ),
                ("name", models.CharField(max_length=50, unique=True)),
                (
                    "rule",
                    models.CharField(
                        max_length=50,
                        validators=[django_fast_iprestrict.validators.validate_network],
                    ),
                ),
                (
                    "action",
                    models.CharField(
                        choices=[("a", "allow"), ("b", "deny"), ("c", "disabled")],
                        default="a",
                        max_length=1,
                    ),
                ),
            ],
            options={
                "ordering": ("position",),
                "indexes": [
                    models.Index(
                        fields=["position"], name="django_fast_positio_d900cb_idx"
                    )
                ],
            },
        ),
        migrations.CreateModel(
            name="RulePath",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("path", models.TextField(max_length=4096)),
                ("is_regex", models.BooleanField(blank=True, default=False)),
                (
                    "rule",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="urls",
                        to="django_fast_iprestrict.rule",
                    ),
                ),
            ],
            options={
                "ordering": ("rule__position", "id"),
            },
        ),
    ]



from django.apps import AppConfig
from django.db.models.signals import post_delete, post_save


def signal_position_cleanup(instance, raw=False, **kwargs):
    from .models import Rule

    if raw:
        return
    if not instance._trigger_cleanup:
        return

    Rule.objects.position_cleanup()


def clear_local_caches(instance, raw=False, **kwargs):
    from .models import Rule

    if raw:
        return

    Rule.objects.clear_local_caches()


class DjangoFastIprestrictConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "django_fast_iprestrict"

    def ready(self):
        from .models import Rule, RuleNetwork, RulePath

        post_delete.connect(
            signal_position_cleanup,
            sender=Rule,
            dispatch_uid="django-fast-iprestrict-after-deletion-rule",
        )

        post_save.connect(
            signal_position_cleanup,
            sender=Rule,
            dispatch_uid="django-fast-iprestrict-after-save-rule",
        )
        post_save.connect(
            clear_local_caches,
            sender=RuleNetwork,
            dispatch_uid="django-fast-iprestrict-after-save-rule-network",
        )
        post_delete.connect(
            clear_local_caches,
            sender=RuleNetwork,
            dispatch_uid="django-fast-iprestrict-after-deletion-rule-network",
        )

        post_save.connect(
            clear_local_caches,
            sender=RulePath,
            dispatch_uid="django-fast-iprestrict-after-save-rule-path",
        )
        post_delete.connect(
            clear_local_caches,
            sender=RulePath,
            dispatch_uid="django-fast-iprestrict-after-deletion-rule-path",
        )



import ipaddress
import re

from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.validators import MinLengthValidator


def validate_network(value):
    try:
        ip_addr = ipaddress.ip_address(value)
        if getattr(ip_addr, "ipv4_mapped", None):
            raise ValidationError(
                "Mapped ip4 addresses are forbidden. Use ip4 address instead.",
                code="invalid",
                params={"value": value},
            )
        # skip further checks, as every ip is also a network
        return
    except ValueError:
        pass
    try:
        ipaddress.ip_network(value, strict=False)
    except ValueError:
        raise ValidationError(
            'Enter a valid IPv4 or IPv6 network or "*".',
            code="invalid",
            params={"value": value},
        )


def validate_regex(value):
    try:
        re.compile(value)
    except re.error:
        raise ValidationError("Invalid regex.", code="invalid", params={"value": value})


min_length_1 = MinLengthValidator(1)


def validate_ratelimit_key(value):
    min_length_1(value)
    splitted = value.split(":", 1)
    if splitted[0] == "django_fast_iprestrict.apply_iprestrict":
        raise ValidationError(
            "ratelimit key would cause infinite recursion",
            code="insecure",
            params={"value": value},
        )
    path_parts = splitted[0].split(".")
    if not all(map(lambda x: x.isidentifier(), path_parts)):
        raise ValidationError("Invalid path.", code="invalid", params={"value": value})
    if path_parts[-1].startswith("_"):
        raise ValidationError(
            "not a safe ratelimit key.", code="insecure", params={"value": value}
        )

    for prefix in getattr(settings, "IPRESTRICT_ALLOWED_FN_PREFIXES", ()):
        if value.startswith(prefix):
            return
    if not value.isidentifier():
        raise ValidationError(
            "not a safe ratelimit key.", code="insecure", params={"value": value}
        )


def validate_generator_fn(value):
    min_length_1(value)
    splitted = value.split(":", 1)
    path_parts = splitted[0].split(".")
    if not all(map(lambda x: x.isidentifier(), path_parts)):
        raise ValidationError("Invalid path.", code="invalid", params={"value": value})
    if path_parts[-1].startswith("_"):
        raise ValidationError(
            "not a safe generate_fn.", code="insecure", params={"value": value}
        )

    for prefix in getattr(settings, "IPRESTRICT_ALLOWED_FN_PREFIXES", ()):
        if value.startswith(prefix):
            return
    raise ValidationError(
        "not a safe generate_fn.", code="insecure", params={"value": value}
    )


_rate = re.compile(r"(\d+)/(\d+)?([smhdw])?")


def validate_rate(value):
    matched = _rate.match(value)
    if not matched:
        raise ValidationError("Invalid rate", code="invalid", params={"value": value})



from asgiref.sync import iscoroutinefunction, sync_to_async
from django.core.exceptions import PermissionDenied
from django.utils.decorators import sync_and_async_middleware

from .utils import RULE_ACTION, get_ip

try:
    import django_fast_ratelimit as ratelimit
except ImportError:
    ratelimit = None


@sync_and_async_middleware
def fast_iprestrict(get_response):
    from .models import RulePath

    # One-time configuration and initialization goes here.
    if iscoroutinefunction(get_response):
        amatch_ip_and_path = sync_to_async(RulePath.objects.match_ip_and_path)

        async def middleware(request):
            action, _, ratelimits = (
                await amatch_ip_and_path(get_ip(request), request.path)
            )[1:]
            if ratelimit:
                for rdict in ratelimits:
                    r = await ratelimit.aget_ratelimit(
                        request=request,
                        action=ratelimit.Action.INCREASE,
                        group=rdict["group"],
                        key=rdict["key"],
                        rate=rdict["rate"],
                    )
                    await r.adecorate_object(
                        request,
                        name=rdict["decorate_name"],
                        block=rdict["block"],
                        wait=rdict["wait"],
                    )
            if action == RULE_ACTION.deny.value:
                raise PermissionDenied()
            response = await get_response(request)
            return response

    else:

        def middleware(request):
            action, _, ratelimits = RulePath.objects.match_ip_and_path(
                get_ip(request), request.path
            )[1:]

            if ratelimit:
                for rdict in ratelimits:
                    r = ratelimit.get_ratelimit(
                        request=request,
                        action=ratelimit.Action.INCREASE,
                        group=rdict["group"],
                        key=rdict["key"],
                        rate=rdict["rate"],
                    )
                    r.decorate_object(
                        request, name=rdict["decorate_name"], block=rdict["block"]
                    )
            if action == RULE_ACTION.deny.value:
                raise PermissionDenied()
            response = get_response(request)
            return response

    return middleware
